/*
** Sybase Open Client 
** Confidential property of Sybase, Inc.
** (c) Copyright Sybase, Inc. 1992, 1996
** All rights reserved.
*/

/*
**  
**  thrdfunc.c 
**
**  Thread and synchronization routines for DCE pthreads. 
**  
**  
*/

#define DEBUG 1

#define	_INCLUDE_POSIX_SOURCE

#if	(dce_sgi || dce_sgi64)
#include	<dce/pthread.h>
#else
#include	<pthread.h>
#endif	/* (dce_sgi || dce_sgi64) */

#include	<stdio.h>
#include	<string.h>
#include	<memory.h>
#include	<sys/errno.h>
#include	<ctpublic.h>

#include 	<thrdfunc.h>


typedef struct _dcesem_
{
	CS_INT		ev;	/* The value of the semaphore */
	pthread_mutex_t pm;	/* Mutex for access to the sem */
	pthread_cond_t  pc;	/* The condition variable */
} DCESEM;

/*
** ex_create_mutex()
**
**	Create a DCE mutex
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_create_mutex(mutex)
CS_VOID	**mutex;
{
	pthread_mutex_t		*pm;
	pthread_mutexattr_t	*pmap;

	if((pm = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t)))
		== (pthread_mutex_t *)NULL)
	{
		return(CS_FAIL);
	}

	if((pmap = (pthread_mutexattr_t *)malloc(sizeof(pthread_mutexattr_t)))
		== (pthread_mutexattr_t *)NULL)
	{
                free(pm);
		return(CS_FAIL);
	}

	if(pthread_mutexattr_create(pmap) < 0)
	{
		free(pm);
		free(pmap);
		return(CS_FAIL);
	}

	if(pthread_mutexattr_setkind_np(pmap, MUTEX_RECURSIVE_NP) < 0)
	{
		free(pm);
		free(pmap);
		return(CS_FAIL);
	}

	if(pthread_mutex_init(pm, *pmap) == -1)
	{
		free(pm);
		free(pmap);
		return(CS_FAIL);
	}

	free(pmap);

	*mutex = (CS_VOID *)pm;
	return(CS_SUCCEED);
}

/*
** ex_delete_mutex()
**
**	Delete a DCE mutex
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_delete_mutex(mutex)
CS_VOID	*mutex;
{
	if(mutex == (CS_VOID *)NULL)
	{
		return(CS_FAIL);
	}

	if(pthread_mutex_destroy((pthread_mutex_t *)mutex) < 0)
	{
		return(CS_FAIL);
	}

	free(mutex);
	return(CS_SUCCEED);
}

/*
** ex_lock_mutux()
**
**	Lock a DCE mutex
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_take_mutex(mutex)
CS_VOID *mutex;
{
	
	if(pthread_mutex_lock((pthread_mutex_t *)mutex) < 0)
	{
		return(CS_FAIL);
	}
	
	return(CS_SUCCEED);
}

/*
** ex_unlock_mutux()
**
**	Unlock a DCE mutex
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_release_mutex(mutex)
CS_VOID *mutex;
{
	
	if(pthread_mutex_unlock((pthread_mutex_t *)mutex) == -1)
	{
		return(CS_FAIL);
	}
	return(CS_SUCCEED);
}

/*
** ex_create_sem()
**
**	Create a DCE semaphore
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_create_sem(sem, count)
CS_VOID **sem;
CS_UINT	count;
{
	DCESEM *dces;

	if((dces = (DCESEM *)malloc(sizeof(DCESEM))) == (DCESEM *)NULL)
	{
		return(CS_FAIL);
	}

	if(pthread_mutex_init(&(dces->pm), pthread_mutexattr_default) < 0)
	{
		free(dces);
		return(CS_FAIL);
	}

	if(pthread_cond_init(&(dces->pc), pthread_condattr_default) == -1)
	{
		free(dces);
		return(CS_FAIL);
	}

	dces->ev = (CS_INT)0;
	*sem = (CS_VOID *)dces;
	return(CS_SUCCEED);
}

/*
** ex_delete_sem()
**
**	Delete a DCE semaphore
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_delete_sem(sem)
CS_VOID *sem;
{
	DCESEM	*dces;

	/* Initialization						*/
	dces = (DCESEM *)sem;

	if(dces == NULL)
	{
		return(CS_FAIL);
	}

	if(pthread_cond_destroy(&(dces->pc)) < 0)
	{
		return(CS_FAIL);
	}

	if(pthread_mutex_destroy(&(dces->pm)) < 0)
	{
		return(CS_FAIL);
	}

	free(dces);
	return(CS_SUCCEED);
}

/*
** ex_release_sem()
**
**	Release a DCE semaphore
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_release_sem(sem)
CS_VOID	*sem;
{
	DCESEM	*dces;

	/* Initialization						*/
	dces = (DCESEM *)sem;

	if(pthread_mutex_lock(&(dces->pm)) < 0)
	{
		return(CS_FAIL);
	}

	/* Set the event						*/
	dces->ev++ ;

	/* We need to only broadcast if we are moving from 0		*/
	if(dces->ev == 1)
	{
		if(pthread_cond_broadcast(&(dces->pc)) == -1)
		{

			if(pthread_mutex_unlock(&(dces->pm)) < 0)
			{
			}
			return(CS_FAIL);
		}
	}

	if(pthread_mutex_unlock(&(dces->pm)) < 0)
	{
		return(CS_FAIL);
	}
	return(CS_SUCCEED);
}

/*
** ex_waitfor_sem()
**
**	Wait on a DCE semaphore. 
**
** Returns:
**	CS_SUCCEED
**	CS_FAIL
**
** Side Effects:
**      None
*/
CS_RETCODE 
ex_waitfor_sem(sem)
CS_VOID	*sem;
{
	DCESEM		*dces;

	/* Initialization						*/
	dces = (DCESEM *)sem;

	if(pthread_mutex_lock(&(dces->pm)) < 0)
	{
		return(CS_FAIL);
	}

	if(dces->ev < 0)
	{
		return(CS_FAIL);
	}

	if(dces->ev > 0)
	{
		dces->ev--;

		if(pthread_mutex_unlock(&(dces->pm)) < 0)
		{
			return(CS_FAIL);
		}
		return(CS_SUCCEED);
	}

	/* Now wait for the condition to be set */

	while(dces->ev == 0)
	{
		if(pthread_cond_wait(&(dces->pc), &(dces->pm)) < 0)
		{

			if(pthread_mutex_unlock(&(dces->pm)) < 0)
			{
			}
			return(CS_FAIL);
		}
	}

	dces->ev--;

	if(pthread_mutex_unlock(&(dces->pm)) == -1)
	{
		return(CS_FAIL);
	}
	return(CS_SUCCEED);
}

/*
**  ex_yield
**  
**  Type of function: internal
**  
**  Purpose:
**      Give up timeslice.
**  
**  Return value:
**  		none
**  
**  Side effects:
**  		none
**  
**  Synchronization:
**  		none
**  
**  Assumptions:
**  		none
**  
**  Notes:
**  		none
**  
**  History:
**  
*/
CS_VOID   
ex_yield()
{
        pthread_yield();
}

/*
**  ex_create_thread
**  
**  Type of function: internal
**  
**  Purpose:
**      allocate the necessary data structures, start a new thread
**  
**  Return value:
**  		CS_FAIL or CS_SUCCEED.
**  
**  Side effects:
**  		none
**  
**  Synchronization:
**  		none
**  
**  Assumptions:
**  		none
**  
**  Notes:
**  		none
**  
**  History:
**  
*/
CS_RETCODE   
ex_create_thread(thread, startRoutine, arg)
CS_VOID                 **thread;
CS_VOID			*startRoutine;
CS_VOID                 *arg;
{
        CS_RETCODE	retstat;
        pthread_t       *pt;
        
        if((pt = (pthread_t *)malloc(sizeof(pthread_t))) == (pthread_t *)NULL)
        {
                return(CS_FAIL);
        }
        if(pthread_create(pt, pthread_attr_default,
                          (pthread_startroutine_t)startRoutine, arg) < 0)
        {                 
		free(pt);
                return(CS_FAIL);
        }
        *thread = pt;
        return(CS_SUCCEED);
}


/*
**
**  ex_detach_thread
**
**  Type of function: external
**
**  Purpose:
**	Detach internal thread resources. Thread object is not anymore
**	accessible.
**
**  Return value:
**              CS_FAIL or CS_SUCCEED.
**
**  Side effects:
**
**  Synchronization:
**              none
**
**  Assumptions:
**
**  Notes:
**
**  History:
**
*/
CS_RETCODE 
ex_detach_thread(threadp)
CS_VOID    *threadp;
{
        pthread_t       *handle;        /* thread to kill               */
        CS_INT          retvalue;

        /* Initialize */
        handle = (pthread_t *)threadp;

        /*
        ** Detach all resources associated with this thread
        */
        (void)pthread_detach(handle);

        /* Get the real Free function  and free the handle */
        free(threadp);

        return CS_SUCCEED;
}

/*
** ex_take_global_mutex()
**
**	Take global mutex
**
** Returns:
**      CS_SUCCEED - always
**
** Side Effects:
**      None
*/
CS_RETCODE   
ex_take_global_mutex()
{
        pthread_lock_global_np();
	return(CS_SUCCEED);
}

/*
** ex_release_global_mutex()
**
**      Release global mutex
**
** Returns:
**      CS_SUCCEED - always
**
** Side Effects:
**      None
*/
CS_RETCODE   
ex_release_global_mutex()
{
        pthread_unlock_global_np();
	return(CS_SUCCEED);
}

/*
** ex_create_global_mutex()
**
**	Setup global mutex mechanism. 
**	For this thread scheme, this is a empty stub routine. 
**
** Returns:
**      CS_SUCCEED - always
**
** Side Effects:
**      None
*/
CS_RETCODE   
ex_create_global_mutex()
{
	return(CS_SUCCEED);
}

/*
** ex_delete_global_mutex()
**
**      Delete global mutex mechanism. 
**      For this thread scheme, this is a empty stub routine.
**
** Returns:
**      CS_SUCCEED - always
**
** Side Effects:
**      None
*/
CS_RETCODE   
ex_delete_global_mutex()
{
	return(CS_SUCCEED);
}


